"""
Action command postprocessor
"""
from chatbots.robot import CompletionReaction, CompletionReactionInterface
from .utils import split_completion, COMMAND_ACTION


class ActionCompletionReaction(CompletionReactionInterface):
    """
    Action command postprocessor
    """
    def check(self, completion: str, variables: dict) -> CompletionReaction:
        for command in split_completion(completion):
            if command.type == COMMAND_ACTION:
                return CompletionReaction(command.args, True)
        return CompletionReaction(None, False)



"""
Logging decorator for search systems
"""
from typing import List
from chatbots.search import SearchInterface
from .base_logger import BaseLogger
from .channels import LoggerChannelInterface


class SearchLogger(SearchInterface, BaseLogger):
    """
    Logging decorator for search systems
    """
    def __init__(self, logger_name: str, logger_channel: LoggerChannelInterface,
                 search: SearchInterface) -> None:
        SearchInterface.__init__(self)
        BaseLogger.__init__(self, logger_name, logger_channel)
        self.search_system = search

    def update(self, document: str, amendment: str) -> None:
        self.search_system.update(document, amendment)
        self.log(
            "Update documents",
            {
                "document": document,
                "amendment": amendment,
            }
        )

    def search(self, query: str) -> List[str]:
        result = self.search_system.search(query)
        self.log(
            "Search request",
            {
                "query": query,
                "result": result,
            }
        )
        return result



"""
Text search through local DB by embeddings similarity
"""
from dataclasses import dataclass
import json
import os
from typing import List, Union, Tuple
import fuzzysearch
from nltk.corpus import stopwords
from nltk.tokenize import wordpunct_tokenize
from .search_interface import SearchInterface
from .utils import is_local_request


@dataclass
class Paragraph:
    """
    Document paragraphs
    """
    text: str
    processed_text: str
    document_name: str


__DEFAULT_SEARCH_PARAMS__ = {
    "max_l_dist": 3,
}


class SearchLocalDatabaseTextual(SearchInterface):
    """
    Semantic search through local DB by embeddings similarity
    """
    def __init__(self, directory: str, language: str, top_results: int,
                 search_params: Union[dict, None] = None) -> None:
        self.directory = directory
        self.language = language
        self.paragraphs = self._scan(self.directory)
        self.stopwords = stopwords.words(language)
        self.top_results = top_results
        if search_params is None:
            search_params = {}
        self.search_params = dict(__DEFAULT_SEARCH_PARAMS__, **search_params)

    def _preprocess(self, text: str) -> str:
        tokens = wordpunct_tokenize(text.lower())
        tokens = filter(str.isalnum, tokens)
        tokens = filter(lambda token: token not in self.stopwords, tokens)
        return " ".join(tokens)

    def _scan_directory_files(self, directory: str) -> List[Tuple[str, str]]:
        paragraph_files = []
        for document_path in os.listdir(directory):
            document_path_full = os.path.join(directory, document_path)
            if not os.path.isdir(document_path_full):
                continue
            for paragraph_path in os.listdir(document_path_full):
                paragraph_path_full = os.path.join(document_path_full, paragraph_path)
                if os.path.isdir(paragraph_path_full):
                    continue
                if not paragraph_path_full.lower().endswith(".txt"):
                    continue
                paragraph_files.append((document_path, paragraph_path_full))
        return paragraph_files

    def _scan(self, directory: str) -> List[Paragraph]:
        paragraphs = []
        documents = self._scan_directory_files(directory)
        for document, filename in documents:
            with open(filename, "r", encoding="utf-8") as src:
                paragraph_json = json.load(src)
                paragraphs.append(Paragraph(
                    text=paragraph_json["text"],
                    processed_text=paragraph_json["processed_text"],
                    document_name=document,
                ))
        return paragraphs

    def update(self, document: str, amendment: str) -> None:
        document_full_path = os.path.join(self.directory, self._preprocess(document))
        if not os.path.exists(document_full_path):
            os.makedirs(document_full_path)

        paragraph_count = len(self._scan_directory_files(self.directory))

        new_paragraphs_text = amendment.split("\n")
        new_paragraphs_text = [item.strip() for item in new_paragraphs_text]
        new_paragraphs_processed = [self._preprocess(item) for item in new_paragraphs_text]
        new_paragraphs = [
            Paragraph(text=text, processed_text=processed_text, document_name=document)
            for text, processed_text in zip(new_paragraphs_text, new_paragraphs_processed)
            if processed_text != ""
        ]

        for i, paragraph in enumerate(new_paragraphs):
            new_paragraph_fname = f"PARAGRAPH_{i + 1 + paragraph_count}.TXT"
            new_paragraph_fname_full = os.path.join(document_full_path, new_paragraph_fname)
            with open(new_paragraph_fname_full, "w", encoding="utf-8") as target:
                json.dump(
                    {
                        "text": paragraph.text,
                        "processed_text": paragraph.processed_text,
                    },
                    target
                )

        self.paragraphs = self._scan(self.directory)

    def search(self, query: str) -> List[str]:
        _, query = is_local_request(query)
        query = self._preprocess(query)
        paragraph_distances = []
        for paragraph in self.paragraphs:
            paragraph_matches = fuzzysearch.find_near_matches(query,
                                                              paragraph.processed_text,
                                                              **self.search_params)
            if len(paragraph_matches) > 0:
                distances = map(lambda match: match.dist, paragraph_matches)
                min_distance = min(distances)
                paragraph_distances.append((min_distance, paragraph))
        paragraph_distances = sorted(paragraph_distances, key=lambda pair: pair[0])
        paragraphs = [
            paragraph
            for _, paragraph in paragraph_distances
        ]
        paragraphs = paragraphs[:self.top_results]
        return [
            paragraph.document_name + " : " + paragraph.text
            for paragraph in paragraphs
        ]



"""
Logger for language model completion processors
"""
from chatbots.robot import CompletionReactionInterface, CompletionReaction
from .base_logger import BaseLogger
from .channels import LoggerChannelInterface


class CompletionReactionLogger(CompletionReactionInterface, BaseLogger):
    """
    Logger for language model completion processors
    """
    def __init__(self, logger_name: str, logger_channel: LoggerChannelInterface,
                 reaction: CompletionReactionInterface) -> None:
        CompletionReactionInterface.__init__(self)
        BaseLogger.__init__(self, logger_name, logger_channel)
        self.reaction = reaction

    def check(self, completion: str, variables: dict) -> CompletionReaction:
        result = self.reaction.check(completion, variables)
        self.log(
            "Check reaction",
            {
                "completion": completion,
                "variables": variables,
                "result": {
                    "answer": result.answer,
                    "stop": result.stop,
                }
            }
        )
        return result



from .search_interface import SearchInterface
from .search_duckduckgo import SearchDuckDuckGo
from .search_local_database_textual import SearchLocalDatabaseTextual
from .search_local_database_semantic import SearchLocalDatabaseSemantic
from .search_ranker import SearchRanker, SearchRankerItem
from .nn_config import NNConfig
from .search_cache import SearchCache



"""
Logger decorators for main robot components
"""
from .channels import LoggerChannelFile, LoggerChannelStderr, LoggerChannelInterface
from .base_logger import BaseLogger
from .language_model_logger import LMLogger
from .search_logger import SearchLogger
from .completion_reaction_logger import CompletionReactionLogger



"""
LM prompt template filler
"""
from typing import Dict
import os


__MODULE_DIR__ = os.path.dirname(__file__)


class PromptFiller:
    """
    LM prompt template filler
    """
    def __init__(self, prompt_fname: str = os.path.join(__MODULE_DIR__, "prompt.txt")) -> None:
        with open(prompt_fname, "r", encoding="utf-8") as src:
            self.prompt = src.read()

    def fill(self, variables: Dict[str, str]) -> str:
        """
        Fill template with given values
        """
        text = self.prompt
        for name, value in variables.items():
            text = text.replace(f"%{name}%", value)
        return text



"""
Semantic search through local DB by embeddings similarity
"""
from dataclasses import dataclass
import json
import os
import pickle
from typing import List, Union, Tuple
import nltk
import numpy as np
from sklearn.neighbors import KNeighborsClassifier
from .nn_config import NNConfig
from .search_interface import SearchInterface
from .utils import get_sentence_transformer, is_local_request


@dataclass
class Paragraph:
    """
    Document paragraphs
    """
    document_name: str
    text: str
    sentence_embeddings: np.ndarray


__DEFAULT_KNN_PARAMS__ = {
    "weights": "distance"
}


class SearchLocalDatabaseSemantic(SearchInterface):
    """
    Semantic search through local DB by embeddings similarity
    """
    def __init__(self, directory: str, model: str, top_n: int, embedder_params: NNConfig,
                 knn_params: Union[dict, None] = None) -> None:
        super().__init__()
        self.embedder_params = embedder_params
        self.model = get_sentence_transformer(model, device=self.embedder_params.device)
        self.encoder_dim = self.model.encode([""]).shape[-1]
        self.directory = directory
        self.top_n = top_n
        if knn_params is None:
            knn_params = {}
        self.knn_params = dict(__DEFAULT_KNN_PARAMS__, **knn_params)
        self.paragraphs, self.knn = self._scan(self.directory, False, self.top_n, self.knn_params)

    def _embeddings_to_string(self, embeddings: np.ndarray) -> str:
        embeddings_bytes = embeddings.astype(np.float32).tobytes()
        embeddings_bytes_ascii = []
        top_bit = b"\x80"[0]
        top_bit_others = b"\x7f"[0]
        for embeddings_byte in embeddings_bytes:
            embeddings_byte_part1 = (embeddings_byte & top_bit) >> 7
            embeddings_byte_part2 = embeddings_byte & top_bit_others
            embeddings_bytes_ascii.append(embeddings_byte_part1)
            embeddings_bytes_ascii.append(embeddings_byte_part2)
        return bytes(embeddings_bytes_ascii).decode("ascii")

    def _embeddings_from_string(self, embeddings_data: str) -> np.ndarray:
        embeddings_bytes = embeddings_data.encode("ascii")
        embeddings_bytes_parsed_values = []
        for i in range(len(embeddings_bytes) // 2):
            embeddings_byte_part1 = embeddings_bytes[i * 2]
            embeddings_byte_part2 = embeddings_bytes[i * 2 + 1]
            embeddings_byte = (embeddings_byte_part1 << 7) | embeddings_byte_part2
            embeddings_bytes_parsed_values.append(embeddings_byte)
        embeddings_bytes_parsed = bytes(embeddings_bytes_parsed_values)
        return np.frombuffer(bytes(embeddings_bytes_parsed), dtype=np.float32).reshape([-1, self.encoder_dim])

    def _scan(self, directory: str, rebuild_knn: bool, top_n: int, knn_params: dict) \
        -> Tuple[List[Paragraph], Union[KNeighborsClassifier, None]]:
        paragraphs = []
        for document_name in os.listdir(directory):
            document_name_full = os.path.join(directory, document_name)
            if os.path.isdir(document_name_full):
                continue
            if not document_name_full.lower().endswith(".json"):
                continue
            with open(document_name_full, "r", encoding="utf-8") as src:
                document_json = json.load(src)
                document_name_original = document_json["name"]
                for paragraph in document_json["paragraphs"]:
                    paragraphs.append(Paragraph(
                        document_name=document_name_original,
                        text=paragraph["text"],
                        sentence_embeddings=self._embeddings_from_string(paragraph["sentence_embeddings"]),
                    ))
        knn_fname = os.path.join(directory, "knn.pkl")
        if (not os.path.exists(knn_fname)) or rebuild_knn:
            knn = self._build_knn(paragraphs, top_n, knn_params)
            with open(knn_fname, "wb") as target:
                pickle.dump(knn, target)
        else:
            with open(knn_fname, "rb") as src:
                knn = pickle.load(src)
        return paragraphs, knn

    def _build_knn(self, paragraphs: List[Paragraph], top_n: int, knn_params: dict) \
        -> Union[KNeighborsClassifier, None]:
        if len(paragraphs) == 0:
            return None
        knn_x = []
        knn_y = []
        for i, paragraph in enumerate(paragraphs):
            for j in range(paragraph.sentence_embeddings.shape[0]):
                knn_x.append(paragraph.sentence_embeddings[j])
                knn_y.append(i)
        knn = KNeighborsClassifier(n_neighbors=top_n, **knn_params)
        knn.fit(np.array(knn_x), np.array(knn_y))
        return knn

    def _process_document_name(self, document: str) -> str:
        return "".join([
            c
            for c in document.lower()
            if c.isalnum()
        ])

    def _encode(self, texts: List[str]) -> np.ndarray:
        return self.model.encode(
            texts,
            batch_size=self.embedder_params.batch_size,
            normalize_embeddings=True
        )

    def update(self, document: str, amendment: str) -> None:
        document_fname = self._process_document_name(document) + ".json"
        document_fname = os.path.join(self.directory, document_fname)
        if os.path.exists(document_fname):
            with open(document_fname, "r", encoding="utf-8") as src:
                document_data = json.load(src)
        else:
            document_data = {
                "name": document,
                "paragraphs": [],
            }

        new_paragraphs = amendment.split("\n")
        new_paragraphs = [item.strip() for item in new_paragraphs]
        new_paragraphs = [item for item in new_paragraphs if item != ""]
        for new_paragraph in new_paragraphs:
            sentences = nltk.sent_tokenize(new_paragraph)
            embeddings = self._encode(sentences)
            document_data["paragraphs"].append({
                "text": new_paragraph,
                "sentence_embeddings": self._embeddings_to_string(embeddings),
            })

        with open(document_fname, "w", encoding="utf-8") as target:
            json.dump(document_data, target)

        self.paragraphs, self.knn = self._scan(self.directory, True, self.top_n, self.knn_params)

    def search(self, query: str) -> List[str]:
        _, query = is_local_request(query)
        if self.knn is None:
            return [""]
        query_embeddings = self._encode([query])
        paragraph_indices_proba = self.knn.predict_proba(query_embeddings)[0]
        top_indices = (-paragraph_indices_proba).argsort()[:self.top_n]
        top_paragraphs_indices = [self.knn.classes_[idx] for idx in top_indices]
        result = []
        for idx in top_paragraphs_indices:
            paragraph = self.paragraphs[idx]
            result.append(f"{paragraph.document_name} : {paragraph.text}")
        return result



"""
Library for chatbot systems backed by:
- language models
- external search
"""
from .botloggers import *
from .completion_reactions import *
from .lm_utils import *
from .robot import *
from .search import *
from .session import RobotConfig, RobotResponse, RobotSession



from dataclasses import dataclass
from typing import List, Dict
from chatbots.robot import Robot, CompletionReactionInterface
from chatbots.search import SearchInterface
from chatbots.lm_utils import LanguageModelInterface, PromptFiller
from chatbots.completion_reactions import SearchCompletionReaction, AnswerCompletionReaction, DiceCompletionReaction, ActionCompletionReaction


@dataclass
class RobotConfig:
    prompt_filler: PromptFiller
    language_model: LanguageModelInterface
    filler_vars: Dict[str, str]


@dataclass
class RobotResponse:
    dialogue: List[str]
    hints: List[str]
    response: str


class RobotSession:
    def __init__(self, robot: RobotConfig, overall_search: SearchInterface, session_search: SearchInterface, max_hints_keep: int, max_dialogue_keep: int) -> None:
        self.robot = Robot(
            robot.prompt_filler,
            robot.language_model,
            robot.filler_vars,
            [
                SearchCompletionReaction(overall_search),
                AnswerCompletionReaction(),
                ActionCompletionReaction(),
                DiceCompletionReaction(),
            ]
        )
        self.session_search = session_search
        self.max_hints_keep = max_hints_keep
        self.max_dialogue_keep = max_dialogue_keep

    def _reaction_wrapper_function(self, reaction_index: int):
        def func(reaction: CompletionReactionInterface):
            self.robot.completion_reactions[reaction_index] = reaction
        
        return func

    def wrap_reactions_loop(self):
        reactions = self.robot.completion_reactions
        for i in range(len(reactions)):
            wrap_function = self._reaction_wrapper_function(i)
            yield reactions[i], wrap_function

    def phrase_cut(self, phrase: str) -> str:
        if ":" in phrase:
            phrase = ":".join(phrase.split(":")[1:]).strip()
        return phrase
    
    def response(self, hints: List[str], dialogue: List[str], retort: str) -> RobotResponse:
        response = self.robot.response(hints, dialogue, retort)
        dialogue.append(retort)
        dialogue.append(response)
        if len(dialogue) > self.max_dialogue_keep:
            dialogue = dialogue[-self.max_dialogue_keep:]
        if len(hints) > self.max_hints_keep:
            hints = hints[-self.max_hints_keep:]
        self.session_search.update("session", "\n".join([self.phrase_cut(retort), self.phrase_cut(response)]))
        return RobotResponse(dialogue, hints, response)



"""
Basic logging channels (stderr / file)
"""
from typing import Dict
import sys


class LoggerChannelInterface:
    """
    Basic "logger channel" interface
    """
    def __init__(self) -> None:
        self.channel_vars = {}

    def get_channel_vars(self) -> Dict[str, str]:
        """
        Get channel-specific variables
        """
        return self.channel_vars

    def set_channel_var(self, name: str, value: str) -> None:
        """
        Set channel-specific variable
        """
        self.channel_vars[name] = value

    def log(self, message: str) -> None:
        """
        Write log message
        """


class LoggerChannelFile(LoggerChannelInterface):
    """
    Write logs to file
    """
    def __init__(self, name: str) -> None:
        super().__init__()
        self.name = name

    def log(self, message: str) -> None:
        with open(self.name, "a", encoding="utf-8") as target:
            target.write(f"{message}\n")
            target.flush()


class LoggerChannelStderr(LoggerChannelInterface):
    """
    Write logs to stderr
    """
    def log(self, message: str) -> None:
        sys.stderr.write(f"{message}\n")
        sys.stderr.flush()



"""
Dice command processor
"""
import random
import re
from chatbots.robot import CompletionReaction, CompletionReactionInterface
from .utils import split_completion, COMMAND_DICE


class DiceCompletionReaction(CompletionReactionInterface):
    """
    Dice command processor
    """
    def check(self, completion: str, variables: dict) -> CompletionReaction:
        for command in split_completion(completion):
            if command.type == COMMAND_DICE:
                items = re.split(r"\n\s*-", command.args)
                items = [item.strip() for item in items]
                items = [item for item in items if item]
                chosen = random.choice(items)
                return CompletionReaction(chosen, True)
        return CompletionReaction(None, False)



"""
Base language model interface
"""

class LanguageModelInterface:
    """
    Base language model interface
    """
    def complete(self, prompt: str) -> str:
        """
        Generate completion
        """
        raise NotImplementedError()



"""
Search command processor
"""
from chatbots.robot import CompletionReaction, CompletionReactionInterface
from chatbots.search import SearchInterface
from .utils import split_completion, COMMAND_SEARCH


class SearchCompletionReaction(CompletionReactionInterface):
    """
    Search command processor
    """
    def __init__(self, search: SearchInterface):
        self.search = search

    def check(self, completion: str, variables: dict) -> CompletionReaction:
        hints = variables["hints"]
        for command in split_completion(completion):
            if command.type != COMMAND_SEARCH:
                continue
            query = command.args
            for query_line in query.split("\n"):
                query_line = query_line.strip()
                query_line_no_story = query_line.replace("STORY", "").strip()
                if not query_line or not query_line_no_story:
                    continue
                query_line_searched = any([
                    hint.startswith(query_line_no_story) or hint.startswith(query_line)
                    for hint in hints
                ])
                if query_line_searched:
                    continue
                query_hints = self.search.search(query_line)
                query_hints = [item.strip() for item in query_hints]
                query_hints = [item for item in query_hints if item]
                for query_hint in query_hints:
                    hints.append(f"{query_line_no_story} - {query_hint}")
        return CompletionReaction(None, stop=False)



"""
Basic language model completion postrocessor
"""
from dataclasses import dataclass
from typing import Union


@dataclass
class CompletionReaction:
    """
    LM output postprocessor result
    """
    answer:  Union[str, None]
    stop: bool


class CompletionReactionInterface:
    """
    Basic language model completion postrocessor
    """
    def check(self, completion: str, variables: dict) -> CompletionReaction:
        """
        Check LM output for this  action
        """
        raise NotImplementedError()



"""
GPT3 API based language model wrapper
"""
from typing import Union
import openai
from .language_model_interface import LanguageModelInterface


__COMPLETION_DEFAULTS__ = {
    "model": "text-davinci-002",
    "temperature": 1.0,
    "max_tokens": 150,
    "frequency_penalty": 1.0,
    "presence_penalty": 1.0,
    "top_p": 1,
    "stop": ["Endseparator:"],
}


class LanguageModelGPT3(LanguageModelInterface):
    """
    GPT3 API based language model wrapper
    """
    def __init__(self, completion_params: Union[dict, None] = None) -> None:
        super(LanguageModelGPT3, self).__init__()
        if completion_params is None:
            completion_params = {}
        self.completion_params = dict(__COMPLETION_DEFAULTS__, **completion_params)

    def complete(self, prompt: str) -> str:
        completion = openai.Completion.create(
            prompt=prompt.strip(),
            **self.completion_params
        )
        if len(completion.choices) == 0:
            return ""
        return completion.choices[0].text.strip().replace("Endseparator:", "").strip()



"""
Cache for internet search
"""
import json
import os
from typing import List
from .search_interface import SearchInterface


# TODO: non-file cache
class SearchCache(SearchInterface):
    """
    Cache for internet search
    """
    def __init__(self, fname: str, search: SearchInterface) -> None:
        super().__init__()
        self.fname = fname
        self.cache = {}
        if os.path.exists(self.fname):
            with open(self.fname, "r", encoding="utf-8") as src:
                self.cache = json.load(src)
        self.search_system = search

    def update(self, document: str, amendment: str) -> None:
        self.search_system.update(document, amendment)

    def search(self, query: str) -> List[str]:
        if query not in self.cache:
            self.cache[query] = self.search_system.search(query)
            with open(self.fname, "w", encoding="utf-8") as target:
                json.dump(self.cache, target)
        return self.cache[query]



from .utils import split_completion, Command
from .search_reaction import SearchCompletionReaction
from .answer_reaction import AnswerCompletionReaction
from .action_reaction import ActionCompletionReaction
from .dice_reaction import DiceCompletionReaction


"""
Search system utils
"""
from typing import Tuple
from sentence_transformers import CrossEncoder, SentenceTransformer


STORY_SEARCH_MARKER = "STORY"


_singletone_crossencoders = {}
_singletone_sentence_transformers = {}


def get_crossencoder(model: str, *args, **kwargs) -> CrossEncoder:
    """
    Get crossencoder instance (one per process)
    """
    if model not in _singletone_crossencoders:
        _singletone_crossencoders[model] = CrossEncoder(model, *args, **kwargs)
    return _singletone_crossencoders[model]


def get_sentence_transformer(model: str, *args, **kwargs) -> SentenceTransformer:
    """
    Get sentence encoder instance (one per process)
    """
    if model not in _singletone_sentence_transformers:
        _singletone_sentence_transformers[model] = SentenceTransformer(model, *args, **kwargs)
    return _singletone_sentence_transformers[model]


def is_local_request(query: str) -> Tuple[bool, str]:
    """
    Check if LM requests search through conversation history or globally
    """
    return STORY_SEARCH_MARKER in query, query.replace(STORY_SEARCH_MARKER, "")



"""
Base search system interface
"""
from typing import List


class SearchInterface:
    """
    Base search system interface
    """
    def search(self, query: str) -> List[str]:
        """
        Perform search through document collection
        """
        raise NotImplementedError()

    def update(self, document: str, amendment: str) -> None:
        """
        Update document collection
        """
        raise NotImplementedError()



"""
Bot main part - prompt filler/language model/LM completions handlers glue
"""
from typing import List, Dict
from chatbots.lm_utils import PromptFiller, LanguageModelInterface
from .completion_reaction_interface import CompletionReactionInterface


__MAX_ITERS__ = 5
__ANSWER_FORCE_COMPLETION__ = "Answer:"


class Robot:
    """
    Bot main part - prompt filler/language model/LM completions handlers glue
    """
    def __init__(self, prompt_filler: PromptFiller, language_model: LanguageModelInterface,
                 filler_vars: Dict[str, str],
                 completion_reactions: List[CompletionReactionInterface],
                 max_iters: int = __MAX_ITERS__) -> None:
        self.prompt_filler = prompt_filler
        self.language_model = language_model
        self.filler_vars = filler_vars
        self.completion_reactions = completion_reactions
        self.max_iters = max_iters

    def _filter_hints(self, hints: List[str], dialogue: List[str]) -> List[str]:
        return [
            hint.strip()
            for hint in hints
            if hint.split(" : ")[-1] not in "\n".join(dialogue)
        ]

    def _join_string_list(self, texts: List[str]) -> str:
        if texts:
            texts = [item.strip() for item in texts]
            return ("\n - " + "\n - ".join(texts)).strip()
        return ""

    def _generate_lm_completion(self, prompt: str, force: bool) -> str:
        if force:
            prompt = prompt.strip() + "\n" + __ANSWER_FORCE_COMPLETION__
        completion = self.language_model.complete(prompt.strip()).strip()
        if force:
            completion = __ANSWER_FORCE_COMPLETION__ + " " + completion
        return completion

    def response(self, hints: List[str], dialogue: List[str], retort: str) -> str:
        """
        Generate robot response
        """
        result = None
        force_answer = False
        previous_completion = None
        for i in range(self.max_iters):
            hints_to_use = self._filter_hints(hints, dialogue)
            filler_vars = dict(
                self.filler_vars,
                **{
                    "HINTS": self._join_string_list(hints_to_use),
                    "DIALOGUE": self._join_string_list(dialogue),
                    "RETORT": retort.strip()
                }
            )
            prompt = self.prompt_filler.fill(filler_vars)
            if i == self.max_iters - 1:
                force_answer = True
            completion = self._generate_lm_completion(prompt, force_answer)
            if completion == "":
                force_answer = True
            elif previous_completion is not None and completion == previous_completion:
                force_answer = True
            previous_completion = completion
            breaker = False
            for reaction in self.completion_reactions:
                reaction_vars = {
                    "hints": hints,
                }
                reaction_response = reaction.check(completion, reaction_vars)
                if reaction_response.answer is not None:
                    result = reaction_response.answer
                if reaction_response.stop:
                    breaker = True
                    break
            if breaker:
                break
        assert result is not None
        return result.strip().strip("-")



"""
Cross-encoder based search ranking to rank results from a few systems
"""
from dataclasses import dataclass
from typing import List
import numpy as np
from torch.nn.functional import softmax
import torch
from .nn_config import NNConfig
from .search_interface import SearchInterface
from .utils import get_crossencoder, is_local_request


@dataclass
class SearchRankerItem:
    """
    Individual search system config
    """
    system: SearchInterface
    weight: float
    updateable: bool


class SearchRanker(SearchInterface):
    """
    Cross-encoder based search ranking to rank results from a few systems
    """
    def __init__(self, ranker_model: str, ranket_model_neutral_label: int,
                 nn_config: NNConfig, search_systems: List[SearchRankerItem],
                 top_n: int) -> None:
        self.nn_config = nn_config
        self.model = get_crossencoder(ranker_model, device=nn_config.device)
        self.neutral_label = ranket_model_neutral_label
        self.search_systems = search_systems
        self.top_n = top_n

    def update(self, document: str, amendment: str) -> None:
        for system in self.search_systems:
            if system.updateable:
                system.system.update(document, amendment)

    def search(self, query: str) -> List[str]:
        _, query = is_local_request(query)
        outputs = []
        output_weights = []
        for system in self.search_systems:
            system_results = system.system.search(query)
            outputs += system_results
            output_weights += [system.weight] * len(system_results)
        query_output_pairs = [
            (item, query)
            for item in outputs
        ]
        with torch.no_grad():
            scores = self.model.predict(query_output_pairs,
                                        batch_size=self.nn_config.batch_size)
            scores = torch.FloatTensor(scores)
            scores = softmax(scores, dim=-1)
            scores = scores.detach().cpu().numpy()
        scores_neutral = scores[:, self.neutral_label]
        scores_neutral *= np.array(output_weights)
        indices = scores_neutral.argsort()[:self.top_n]
        return [outputs[idx] for idx in indices]



"""
Logger base class module
"""
import datetime
import json
from .channels import LoggerChannelInterface


class BaseLogger:
    """
    Basic logger mixin
    """
    def __init__(self, name: str, channel: LoggerChannelInterface) -> None:
        self.name = name
        self.channel = channel

    def log(self, message: str, params: dict) -> None:
        """
        Log message
        """
        now = str(datetime.datetime.now())
        params_all = dict(self.channel.get_channel_vars(), **params)
        message = f"[{self.name}] {now}: {message} {json.dumps(params_all, ensure_ascii=False)}"
        self.channel.log(message)



"""
Logger decorator for language models
"""
from chatbots.lm_utils import LanguageModelInterface
from .channels import LoggerChannelInterface
from .base_logger import BaseLogger


class LMLogger(LanguageModelInterface, BaseLogger):
    """
    Logger decorator for language models
    """
    def __init__(self, logger_name: str, logger_channel: LoggerChannelInterface,
                 language_model: LanguageModelInterface) -> None:
        LanguageModelInterface.__init__(self)
        BaseLogger.__init__(self, logger_name, logger_channel)
        self.language_model = language_model

    def complete(self, prompt: str) -> str:
        result = self.language_model.complete(prompt)
        self.log(
            "Language model completion",
            {
                "prompt": prompt,
                "response": result
            }
        )
        return result



"""
LM postprocessor utils
"""
import re
from typing import List
from dataclasses import dataclass


COMMAND_SEARCH = "Search"
COMMAND_ANSWER = "Answer"
COMMAND_ACTION = "Action"
COMMAND_DICE = "Dice"
COMMANDS = [
    COMMAND_SEARCH,
    COMMAND_ACTION,
    COMMAND_ANSWER,
    COMMAND_DICE,
]
COMMANDS_WITH_SUFFIX = {
    cmd + ":": cmd
    for cmd in COMMANDS
}


@dataclass
class Command:
    """
    Command extracted from LM completion
    """
    type: str
    args: str


def split_completion(completion: str) -> List[Command]:
    """
    Split LM completion to a sequence of commands
    """
    pattern = "(" + \
        "|".join([
            command + r"\:"
            for command in COMMANDS
        ]) + \
        ")"
    parts = re.split(pattern, completion)
    result = []
    buffer = []
    command = None
    for item in parts:
        if item in COMMANDS_WITH_SUFFIX:
            result.append((command, buffer))
            command = COMMANDS_WITH_SUFFIX[item]
            buffer = []
        else:
            buffer.append(item)
    if buffer:
        result.append((command, buffer))
    result = [
        (command, "".join(items).strip())
        for command, items in result
    ]
    result = [
        (command, items)
        for command, items in result
        if (command is not None) and (items != "")
    ]
    return [
        Command(command, items)
        for command, items in result
    ]



"""
Neural network-based search method run config
"""
from dataclasses import dataclass


@dataclass
class NNConfig:
    """
    Neural network-based search method run config
    """
    device: str
    batch_size: int



import os
from setuptools import setup, find_packages


with open(os.path.join(os.path.dirname(__file__), "requirements.txt"), "r", encoding="utf-8") as req:
    requirements = req.read().strip().splitlines()
with open(os.path.join(os.path.dirname(__file__), "README.md"), "r", encoding="utf-8") as readme:
    readme_text = readme.read()
setup(
    name="auglm-chatbots",
    version="0.0.3",
    packages=find_packages(include=["chatbots", "chatbots.*"]),
    install_requires=requirements,
    author="Alexander Pozharskii",
    author_email="gaussmake@gmail.com",
    url="https://github.com/alex4321/gpt3-retrieving-bot",
    data_files=[
        ('', ['README.md', 'requirements.txt'])
    ],
    include_package_data=True,
    long_description=readme_text,
    long_description_content_type='text/markdown',
)


"""
DuckDuckGo-based search through internet
"""
import time
from typing import List
import urllib.parse
from bs4 import BeautifulSoup
import requests
from .search_interface import SearchInterface
from .utils import is_local_request


__TIMEOUT__ = 1.0
__REQUESTS_TIMEOUT__ = 60.0
__DUCKDUCKGO_SEARCH_URL__ = "https://duckduckgo.com/html/?q={query}"
__FIREFOX_UA__ = 'Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:52.0) Gecko/20100101 Firefox/52.0'


class SearchDuckDuckGo(SearchInterface):
    """
    DuckDuckGo-based search through internet
    """
    def __init__(self, keep_top_n: int) -> None:
        self.keep_top_n = keep_top_n

    def search(self, query: str) -> List[str]:
        local_request, query = is_local_request(query)
        if local_request:
            return []
        url = __DUCKDUCKGO_SEARCH_URL__.format(
            query=urllib.parse.quote(query)
        )
        headers = {
            'Access-Control-Allow-Origin': '*',
            'Access-Control-Allow-Methods': 'GET',
            'Access-Control-Allow-Headers': 'Content-Type',
            'Access-Control-Max-Age': '3600',
            'User-Agent': __FIREFOX_UA__,
        }
        request = requests.get(url, headers, timeout=__REQUESTS_TIMEOUT__)
        document = BeautifulSoup(request.content, "html.parser")
        results = document.select("#links .results_links")
        snippets = [
            result.select_one(".result__snippet").text
            for result in results
        ]
        time.sleep(__TIMEOUT__)
        return snippets[:self.keep_top_n]

    def update(self, document: str, amendment: str) -> None:
        pass
